import flask

from flaskel import httpcode
from flaskel.views import Resource

resources = [
    {'item': 1},
    {'item': 2},
    {'item': 3},
]

sub_resources = [
    {'sub': 1}
]


class APIResource(Resource):
    def on_get(self, res_id):
        try:
            return resources[res_id - 1]
        except IndexError:
            flask.abort(httpcode.NOT_FOUND)

    def on_collection(self):
        return resources

    def on_post(self):
        resources.append(flask.request.get_json())
        return flask.request.get_json(), httpcode.CREATED

    def on_delete(self, res_id):
        resources.pop(res_id)

    def on_put(self, res_id):
        resources[res_id - 1] = flask.request.get_json()
        return resources[res_id - 1]

    def sub_items(self, res_id):
        try:
            return sub_resources[res_id - 1]
        except IndexError:
            return []

    def sub_items_post(self, res_id):
        sub_resources[res_id - 1] = flask.request.get_json()
        return sub_resources[res_id - 1], httpcode.CREATED



import enum


class HashableDict(dict):
    def __hash__(self):
        """

        :return:
        """
        return hash(tuple(sorted(self.items())))


class ObjectDict(dict):
    def __init__(self, **kwargs):
        super().__init__()
        for k, v in kwargs.items():
            self[k] = self.normalize(v)

    def __dict__(self):
        data = {}
        for k, v in self.items():
            if isinstance(v, ObjectDict):
                data[k] = v.__dict__()
            elif isinstance(v, list):
                data[k] = [i.__dict__() for i in v]
            else:
                data[k] = v

        return data

    def __getstate__(self):
        return self.__dict__()

    def __setstate__(self, state):
        for k, v in state.items():
            self.__setattr__(k, v)

    def __getattr__(self, name):
        if name in self:
            return self[name]

    def __setattr__(self, name, value):
        self[name] = self.normalize(value)

    def __delattr__(self, name):
        if name in self:
            del self[name]

    @staticmethod
    def normalize(data):
        try:
            if isinstance(data, (list, tuple, set)):
                return [
                    ObjectDict(**r) if isinstance(r, dict) else r
                    for r in data
                ]
            return ObjectDict(**data)
        except (TypeError, AttributeError):
            return data


class IntEnum(enum.IntEnum):
    @classmethod
    def to_list(cls):
        return [
            ObjectDict(id=getattr(cls, m).value, label=getattr(cls, m).name)
            for m in cls.__members__
        ]

    def to_dict(self):
        return ObjectDict(id=self.value, label=self.name)

    def __repr__(self):
        return f"<{self.value}: {self.name}>"

    def __str__(self):
        return self.name



import os

import pytest

from flaskel import datastruct, middlewares, tester
from flaskel.ext import (
    auth, BASE_EXTENSIONS, caching, crypto, healthcheck,
    ip_ban, jobs, limiter, sqlalchemy, useragent
)
from tests.blueprints import BLUEPRINTS

BASE_DIR = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
SKEL_DIR = os.path.join(BASE_DIR, 'skeleton')
CONF_DIR = os.path.join(SKEL_DIR, 'config')

CTS = datastruct.ObjectDict(
    json='application/json',
    xml='application/xml',
    html='text/html',
    json_problem='application/problem+json',
    xml_problem='application/problem+xml',
    json_health='application/health+json',
)

HOSTS = datastruct.ObjectDict(
    apitester="http://httpbin.org",
    fake="http://localhost"
)

SCHEMA_DIR = 'skeleton/blueprints/api/static/schemas'

SCHEMAS = dict(
    JSONRPC=f"file://{SCHEMA_DIR}/jsonrpc.json",
    APIPROBLEM=f"file://{SCHEMA_DIR}/apiproblem.json",
    HEALTHCHECK=f"file://{SCHEMA_DIR}/healthcheck.json",
    OPENAPI3=f"file://{SCHEMA_DIR}/openapi3.json",
)


@pytest.fixture(scope='session')
def app_prod():
    # noinspection PyUnusedLocal
    def test_health_true(**kwargs):
        return True, None

    # noinspection PyUnusedLocal
    def test_health_false(**kwargs):
        return False, "error"

    healthcheck.health_checks.register('system')(healthcheck.health_system)
    healthcheck.health_checks.register('mongo', db=sqlalchemy.db)(healthcheck.health_mongo)
    healthcheck.health_checks.register('redis', db=sqlalchemy.db)(healthcheck.health_redis)
    healthcheck.health_checks.register('sqlalchemy', db=sqlalchemy.db)(healthcheck.health_sqlalchemy)

    return tester.TestClient.get_app(
        conf=dict(
            DEBUG=True,
            BASIC_AUTH_USERNAME='username',
            BASIC_AUTH_PASSWORD='password',
            USER_AGENT_AUTO_PARSE=True,
            PREFERRED_URL_SCHEME='https',
            SCHEMAS=SCHEMAS,
            USE_X_SENDFILE=True,
            JSONRPC_BATCH_MAX_REQUEST=2,
            PROXIES=dict(
                CONF=dict(
                    host=HOSTS.apitester,
                    url='/anything',
                    headers={'k': 'v'},
                    params={'k': 'v'}
                )
            )
        ),
        blueprints=(
            *BLUEPRINTS,
            *(None,),  # NB: needed to complete coverage
            *((None,),)  # NB: needed to complete coverage
        ),
        extensions={
            **BASE_EXTENSIONS,
            **{
                "empty":         (None, (None,)),  # NB: needed to complete coverage
                "useragent":     (useragent.UserAgent(),),
                "argon2":        (crypto.Argon2(),),
                "jwt":           (auth.jwtm,),
                "limiter":       (limiter,),
                "caching":       (caching,),
                "sqlalchemy":    (sqlalchemy.db,),
                "scheduler":     (jobs.scheduler,),
                "ipban":         (ip_ban, dict(nuisances=dict(string=["/phpmyadmin"]))),
                "health_checks": (
                    healthcheck.health_checks, {'extensions': (
                        {
                            'name': 'health_true',
                            'func': test_health_true,
                        },
                        {
                            'func': test_health_false,
                        },
                    )},
                ),
            }
        },
        middlewares=(
            middlewares.ForceHttps,
            middlewares.HTTPMethodOverride,
            middlewares.ReverseProxied,
            middlewares.RequestID,
        ),
        folders=["skeleton/blueprints/web/templates"],
        static_folder="skeleton/blueprints/web/static"
    )


@pytest.fixture(scope='session')
def app_dev():
    return tester.TestClient.get_app(
        conf=dict(
            DEBUG=True,
            WSGI_WERKZEUG_LINT_ENABLED=True,
            WSGI_WERKZEUG_PROFILER_ENABLED=True,
            FLASK_ENV='development',
            SCHEMAS=SCHEMAS
        ),
        blueprints=BLUEPRINTS,
        extensions=BASE_EXTENSIONS,
        template_folder="skeleton/blueprints/web/templates",
        static_folder="skeleton/blueprints/web/static"
    )


@pytest.fixture(scope='session')
def testapp(app_prod):
    return app_prod.test_client()



import flask
from flask_response_builder import encoders

from . import cap, httpcode
from .utils.datastruct import ObjectDict


class Request(flask.Request):
    @property
    def id(self):
        hdr = cap.config.REQUEST_ID_HEADER
        if hasattr(flask.g, 'request_id'):
            return flask.g.request_id
        elif hdr in flask.request.headers:
            return flask.request.headers[hdr]

        flask_header_name = f"HTTP_{hdr.upper().replace('-', '_')}"
        return flask.request.environ.get(flask_header_name)

    def get_json(self, allow_empty=False):
        """

        :param allow_empty:
        :return:
        """
        payload = super().get_json()
        if payload is None:
            if not allow_empty:
                flask.abort(httpcode.BAD_REQUEST, 'No JSON in request')
            payload = ObjectDict()  # pragma: no cover

        return ObjectDict.normalize(payload)


class Response(flask.Response):
    @staticmethod
    def no_content(status=httpcode.NO_CONTENT, headers=None):
        """

        :param status:
        :param headers:
        :return:
        """
        response = flask.make_response(bytes())
        response.headers.update(headers or {})
        response.headers.pop('Content-Type', None)
        response.headers.pop('Content-Length', None)
        response.status_code = status
        return response

    @staticmethod
    def send_file(directory, filename, **kwargs):
        """

        :param directory:
        :param filename:
        :param kwargs:
        """
        kwargs.setdefault('as_attachment', True)
        file_path = flask.safe_join(directory, filename)

        try:
            resp = flask.send_file(file_path, **kwargs)
        except IOError as exc:
            cap.logger.warning(str(exc))
            flask.abort(httpcode.NOT_FOUND)
            return  # only to prevent warning # pragma: no cover

        # following headers works with nginx compatible proxy
        if cap.use_x_sendfile is True and cap.config.ENABLE_ACCEL is True:
            resp.headers['X-Accel-Redirect'] = file_path
            resp.headers['X-Accel-Charset'] = cap.config.ACCEL_CHARSET or 'utf-8'
            resp.headers['X-Accel-Buffering'] = 'yes' if cap.config.ACCEL_BUFFERING else 'no'
            if cap.config.ACCEL_LIMIT_RATE:
                resp.headers['X-Accel-Limit-Rate'] = cap.config.ACCEL_LIMIT_RATE
            if cap.config.SEND_FILE_MAX_AGE_DEFAULT:
                resp.headers['X-Accel-Expires'] = cap.config.SEND_FILE_MAX_AGE_DEFAULT

        return resp

    def get_json(self):
        return ObjectDict.normalize(super().get_json())


class Flaskel(flask.Flask):
    request_class = Request
    response_class = Response
    json_encoder = encoders.JsonEncoder



import asyncio
import threading

import nest_asyncio


class BatchExecutor:
    def __init__(self, tasks=None):
        """

        :param tasks:
        """
        self._tasks = tasks or []

    @staticmethod
    def prepare_task(task):
        """

        :param task:
        :return:
        """
        try:
            if len(task) > 1:
                return task[0], task[1] or {}
            else:
                return task[0], {}
        except TypeError:
            return task, {}

    def run(self):
        responses = []
        for task in self._tasks:
            func, args = self.prepare_task(task)
            responses.append(func(**args))
        return responses


class AsyncBatchExecutor(BatchExecutor):
    def __init__(self, return_exceptions=False, **kwargs):
        """

        :param tasks:
        :param return_exceptions:
        """
        super().__init__(**kwargs)
        self._return_exceptions = return_exceptions

    @staticmethod
    async def _executor(fun, **kwargs):
        return fun(**kwargs)  # pragma: no cover

    @staticmethod
    def is_async(fun):
        return asyncio.iscoroutinefunction(fun)

    async def batch(self):
        tasks = []
        for task in self._tasks:
            func, args = self.prepare_task(task)
            if not self.is_async(func):
                tasks.append(self._executor(func, **args))  # pragma: no cover
            else:
                tasks.append(func(**args))

        return await asyncio.gather(*tasks, return_exceptions=self._return_exceptions)

    def run(self):
        try:
            asyncio.get_running_loop()
            nest_asyncio.apply()  # pragma: no cover
        except RuntimeError:
            asyncio.set_event_loop(asyncio.new_event_loop())

        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.batch())


class Thread(threading.Thread):
    def __init__(self, runnable, params=None, *args, **kwargs):
        """

        :param fun:
        :param params:
        """
        super().__init__(**kwargs)
        self._runnable = runnable
        self._args = args
        self._params = params or {}
        self.response = None

    def run(self):
        self.response = self._runnable(*self._args, **self._params)


class DaemonThread(threading.Thread):
    def __init__(self, runnable, params=None, *args, **kwargs):
        """

        :param runnable:
        :param params:
        """
        super().__init__(**kwargs)
        self.daemon = True
        self._runnable = runnable
        self._args = args
        self._kwargs = params or {}

    def run(self):
        self._runnable(*self._args, **self._kwargs)


class ThreadBatchExecutor(BatchExecutor):
    def __init__(self, thread_class=None, single_thread=False, **kwargs):
        """

        :param thread_class:
        """
        super().__init__(**kwargs)
        self._single_thread = single_thread
        self._thread_class = thread_class or Thread

        if self._single_thread:
            self._tasks[0] = self._thread_class(super().run)
            return

        for i, t in enumerate(self._tasks):
            if isinstance(t, dict):
                thread = self._thread_class(**t)
            else:
                func, args = self.prepare_task(t)
                thread = self._thread_class(func, params=args)
            self._tasks[i] = thread

    def run(self):
        daemon = False

        for t in self._tasks:
            t.start()

        for t in self._tasks:
            if t.daemon is False:
                t.join()
            else:
                daemon = True

        if daemon is True:
            return

        if self._single_thread:
            return self._tasks[0].response

        return [t.response for t in self._tasks]



import uuid

from flaskel import cap


def get_uuid(ver=4, hexify=True, ns=None, name=None):
    """

    :param ver:
    :param hexify:
    :param ns:
    :param name:
    :return:
    """
    _uuid = None

    if ver == 1:
        _uuid = uuid.uuid1()
    elif ver == 3:
        _uuid = uuid.uuid3(
            ns or uuid.NAMESPACE_DNS, name or cap.config['SERVER_NAME']
        )
    elif ver == 4:
        _uuid = uuid.uuid4()
    elif ver == 5:
        _uuid = uuid.uuid5(
            ns or uuid.NAMESPACE_DNS, name or cap.config['SERVER_NAME']
        )

    return _uuid.hex if hexify else _uuid


def check_uuid(u: str, ver=4, exc=False):
    """

    :param u:
    :param ver:
    :param exc:
    :return:
    """
    try:
        if isinstance(u, uuid.UUID):
            return True

        _uuid = uuid.UUID(u, version=ver)
        return u == (str(_uuid) if '-' in u else _uuid.hex)
    except (ValueError, TypeError, AttributeError):
        if exc:
            raise ValueError(f"'{u}' is an invalid UUID{ver}")
        return False



from twisted.internet import reactor
from twisted.web.server import Site
from twisted.web.wsgi import WSGIResource

from .base import BaseApplication


class WSGITwisted(BaseApplication):
    def __init__(self, app, options=None):
        """

        :param app:
        :param options:
        """
        BaseApplication.__init__(self, app, options)
        resource = WSGIResource(reactor, reactor.getThreadPool(), self.application)
        self._site = Site(resource)

    def run(self):
        """

        """
        reactor.listenTCP(self._port, self._site, interface=self._interface)
        reactor.run()



import pytest

MARKERS = {
    "smoke": {
        "description": "smoke tests",
        "marker":      pytest.mark.smoke
    },
    "email": {
        "description": "email tests",
        "marker":      pytest.mark.email
    },
    "integ": {
        "description": "integration tests",
        "marker":      pytest.mark.integ
    },
    "infra": {
        "description": "infrastructure tests",
        "marker":      pytest.mark.infra
    },
}


def pytest_configure(config):
    """

    :param config:
    """
    for k, v in MARKERS.items():
        config.addinivalue_line(
            "markers", "{}: {}".format(k, v['description'])
        )



import sys

import pytest


def main():
    pytest.main([
        '-v', '-rf', '--strict-markers',
        '-p', 'flaskel.tester.plugins.fixtures',
        '-p', 'flaskel.tester.plugins.startup',
        *sys.argv[1:]
    ])


if __name__ == '__main__':
    main()



import importlib
import string
from random import SystemRandom


def import_from_module(name):
    """

    :param name:
    :return:
    """
    mod, attr = name.split(':')
    module = importlib.import_module(mod)
    return getattr(module, attr)


def parse_value(v):
    """

    :param v:
    :return:
    """
    try:
        return float(v) if '.' in v else int(v)
    except ValueError:
        if v.lower() in ("true", "false"):
            return v.lower() == "true"
    return v


def get_from_dict(d, k, default=None):
    """
    get a value from dict without raising exceptions

    :param d: dict
    :param k: key
    :param default: default value if k is missing
    :return: d[k] or default
    """
    if isinstance(d, dict):
        return d.get(k, default)
    else:
        return default


def to_int(n):
    """
    cast input to int if an error occurred returns None

    :param n: input
    :return: int or None
    """
    try:
        return int(n)
    except (TypeError, ValueError):
        return None


def to_float(n):
    """
    cast input to float if an error occurred returns None

    :param n: input
    :return: float or None
    """
    try:
        return float(n)
    except (TypeError, ValueError):
        return None


def random_string(length, alphabet=string.printable):
    """

    :param length:
    :param alphabet:
    :return:
    """
    return ''.join(
        SystemRandom().choice(alphabet)
        for _ in range(length)
    )



import functools

import flask

from flaskel import cap, httpcode
from flaskel.ext import builder
from flaskel.utils.batch import ThreadBatchExecutor


class HealthCheck:
    def __init__(self, app=None, **kwargs):
        """

        :param app:
        """
        self._executor = None
        self._health_checks = {}

        if app:
            self.init_app(app, **kwargs)  # pragma: no cover

    def init_app(self, app, bp=None, extensions=(), decorators=(), executor=None):
        """

        :param app:
        :param bp:
        :param extensions:
        :param decorators:
        :param executor:
        """
        self._executor = executor or ThreadBatchExecutor
        app.config.setdefault('HEALTHCHECK_ABOUT_LINK', None)
        app.config.setdefault('HEALTHCHECK_PATH', '/healthcheck')
        app.config.setdefault('HEALTHCHECK_PARAM_KEY', 'checks')
        app.config.setdefault('HEALTHCHECK_PARAM_SEP', '+')
        app.config.setdefault('HEALTHCHECK_CONTENT_TYPE', 'application/health+json')

        if not hasattr(app, 'extensions'):
            app.extensions = dict()  # pragma: no cover
        app.extensions['healthcheck'] = self

        view = self.perform
        for decorator in decorators:
            view = decorator(view)

        blueprint = bp or app
        blueprint.add_url_rule(app.config['HEALTHCHECK_PATH'], view_func=view)

        for ex in extensions:
            self.register(**ex)(ex.pop('func'))

    @builder.response('json')
    def perform(self):
        """

        :return:
        """
        healthy = True
        response = dict(
            status='pass', checks={},
            links={"about": cap.config['HEALTHCHECK_ABOUT_LINK']}
        )

        params = flask.request.args.get(cap.config['HEALTHCHECK_PARAM_KEY'])
        all_checks = set(self._health_checks.keys())
        if not params:
            params = all_checks
        else:
            params = params.split(cap.config['HEALTHCHECK_PARAM_SEP'])
            params = set(params).intersection(all_checks)

        # noinspection PyProtectedMember
        app = cap._get_current_object()
        params = list(params)
        tasks = [(self._health_checks.get(p), dict(app=app)) for p in params]
        resp = self._executor(tasks=tasks).run()

        for i, p in enumerate(params):
            state, message = resp[i]
            status = 'pass' if state else 'fail'
            response['checks'][p] = dict(status=status, output=message)
            if not state:
                healthy = False

        if not healthy:
            response['status'] = 'fail'

        status = httpcode.SUCCESS if healthy else httpcode.SERVICE_UNAVAILABLE
        return response, status, {'Content-Type': cap.config['HEALTHCHECK_CONTENT_TYPE']}

    def register(self, name=None, **kwargs):
        """

        :param name:
        :param kwargs:
        :return:
        """

        def _register(func):
            """

            :param func:
            :return:
            """
            nonlocal name
            name = name or func.__name__

            @functools.wraps(func)
            def wrapped():
                self._health_checks[name] = functools.partial(func, **kwargs)

            return wrapped()

        return _register



class BaseApplication:
    default_host = '127.0.0.1'
    default_port = 5000

    def __init__(self, app, options=None):
        """

        :param app:
        :param options:
        """
        self.application = app
        self.options = options or {}

        default_bind = '{}:{}'.format(self.default_host, self.default_port)
        bind = (self.options.get('bind') or default_bind).split(':')

        self._interface = bind[0] or self.default_host
        self._port = int(bind[1]) if len(bind) > 1 else self.default_port

    def run(self):
        """

        :return:
        """
        raise NotImplemented  # pragma: no cover


class WSGIBuiltin(BaseApplication):
    def run(self):
        """

        :return:
        """
        debug = self.application.config.get('DEBUG') or False

        self.application.run(
            host=self._interface,
            port=self._port,
            debug=debug
        )



# noinspection PyUnresolvedReferences
from flask import current_app as cap

from flaskel.http import httpcode
from flaskel.tester import TestClient
from flaskel.utils import datastruct, datetime, JSONSchema, SCHEMAS, uuid, webargs, yaml
from .builder import AppBuilder
from .ext import BASE_EXTENSIONS
from .flaskel import Flaskel, Request, Response
from .standalone import Server
from .version import *
from .wsgi import WSGIFactory



from flask_cloudflare_remote import CloudflareRemote
from flask_cors import CORS
from flask_errors_handler import ErrorHandler
from flask_logify import FlaskLogging
from flask_response_builder import ResponseBuilder
from flask_template_support import TemplateSupport

cors = CORS()
logger = FlaskLogging()
errors = ErrorHandler()
builder = ResponseBuilder()
template = TemplateSupport()
cfremote = CloudflareRemote()



from datetime import datetime

from flaskel import cap


class Millis:
    seconds = 1000
    minute = seconds * 60
    hour = minute * 60
    day = hour * 24


class Seconds:
    millis = 1000
    minute = 60
    hour = minute * 60
    day = hour * 24


class Minutes:
    seconds = 60
    hour = 60
    day = hour * 24


class Day:
    hours = 24
    minutes = hours * 60
    seconds = minutes * 60


def from_iso_format(str_date, fmt, exc=True):
    """

    :param str_date: input date with ISO format
    :param fmt: format input date
    :param exc: raise or not exception (default True)
    :return: return formatted date
    """
    if not (exc or str_date):
        return None  # pragma: no cover

    try:
        date_time = datetime.strptime(str_date, cap.config.DATE_ISO_FORMAT)
        return date_time.strftime(fmt)
    except (ValueError, TypeError):
        if exc is True:
            raise  # pragma: no cover


def to_iso_format(str_date, fmt, exc=True):
    """

    :param str_date: input date
    :param fmt: format input date
    :param exc: raise or not exception (default True)
    :return: return date with ISO format
    """
    if not (exc or str_date):
        return None  # pragma: no cover

    try:
        date_time = datetime.strptime(str_date, fmt)
        return date_time.strftime(cap.config.DATE_ISO_FORMAT)
    except (ValueError, TypeError):
        if exc is True:
            raise  # pragma: no cover



from user_agents import parsers

from flaskel.utils.datastruct import ObjectDict


class UserAgent(parsers.UserAgent):
    def __init__(self, ua_string=None):
        """

        :param ua_string:
        """
        self._parsed = False
        self._cached = None
        self.os = None
        self.device = None
        self.browser = None
        self.ua_string = ua_string

    def parse(self, ua_string=None):
        """

        :param ua_string:
        """
        if ua_string:
            self._cached = None
            self._parsed = False
            self.ua_string = ua_string

        if not self._parsed:
            self._parsed = True
            parsed = parsers.user_agent_parser.Parse(self.ua_string)
            self.os = parsers.parse_operating_system(**parsed['os'])
            self.browser = parsers.parse_browser(**parsed['user_agent'])
            self.device = parsers.parse_device(**parsed['device'])

        return self.to_dict()

    def to_dict(self):
        """

        :return:
        """
        if self._cached:
            return self._cached  # pragma: no cover

        self._cached = ObjectDict(
            raw=self.ua_string,
            browser=dict(
                family=self.browser.family,
                version=dict(
                    number=self.browser.version,
                    string=self.browser.version_string
                )
            ),
            os=dict(
                family=self.os.family,
                version=dict(
                    number=self.os.version,
                    string=self.os.version_string
                )
            ),
            device=dict(
                family=self.device.family,
                brand=self.device.brand,
                model=self.device.model,
                type=dict(
                    mobile=self.is_mobile,
                    tablet=self.is_tablet,
                    pc=self.is_pc,
                    bot=self.is_bot,
                    email_client=self.is_email_client,
                    touch_capable=self.is_touch_capable
                )
            ),
        )
        return self._cached



import re

import jsonschema

from flaskel.utils.schemas import JSONSchema


class BaseAssert:
    assert_fail_message = "Test that {that} failed: got {actual}, expected {expected}"

    @classmethod
    def assert_that(cls, func, actual, expected=None, that="", error=None):
        """

        :param func:
        :param actual:
        :param expected:
        :param that:
        :param error:
        """
        assert func(actual, expected), \
            error or cls.assert_fail_message.format(
                that=that, actual=actual, expected=expected
            )

    @classmethod
    def assert_true(cls, actual, error=None):
        cls.assert_that(
            lambda a, _: bool(a) is True,
            error=error, that=f"'{actual}' is True", actual=actual
        )

    @classmethod
    def assert_false(cls, actual, error=None):
        cls.assert_that(
            lambda a, _: bool(a) is False,
            error=error, that=f"'{actual}' is False", actual=actual
        )

    @classmethod
    def assert_none(cls, actual, error=None):
        cls.assert_that(
            lambda a, _: a is None,
            error=error, that=f"'{actual}' is None", actual=actual
        )

    @classmethod
    def assert_equals(cls, actual, expected, error=None):
        def _equals(a, e):
            if type(a) in (list, tuple):
                a = len(a)
            return a == e

        cls.assert_that(
            lambda a, e: _equals(a, e),
            error=error, that=f"'{actual}' is equals to '{expected}'",
            actual=actual, expected=expected
        )

    @classmethod
    def assert_different(cls, actual, expected, error=None):
        def _different(a, e):
            if type(a) in (list, tuple):
                a = len(a)
            return a != e

        cls.assert_that(
            lambda a, e: _different(a, e),
            error=error, that=f"'{actual}' is different to '{expected}'",
            actual=actual, expected=expected
        )

    @classmethod
    def assert_in(cls, actual, expected, error=None):
        cls.assert_that(
            lambda a, e: a in e,
            error=error, that=f"'{actual}' is in '{expected}'",
            actual=actual, expected=expected
        )

    @classmethod
    def assert_not_in(cls, actual, expected, error=None):
        cls.assert_that(
            lambda a, e: a not in e,
            error=error, that=f"'{actual}' is not in '{expected}'",
            actual=actual, expected=expected
        )

    @classmethod
    def assert_range(cls, actual, expected, error=None):
        cls.assert_that(
            lambda a, e: e[0] <= a <= e[1],
            error=error, that=f"'{actual}' is in ranger '{expected}'",
            actual=actual, expected=expected
        )

    @classmethod
    def assert_greater(cls, actual, expected, error=None):
        def _greater(a, e):
            if type(a) in (list, tuple):
                a = len(a)
            return a > e

        cls.assert_that(
            lambda a, e: _greater(a, e),
            error=error, that=f"'{actual}' is greater than '{expected}'",
            actual=actual, expected=expected
        )

    @classmethod
    def assert_less(cls, actual, expected, error=None):
        def _less(a, e):
            if type(a) in (list, tuple):
                a = len(a)
            return a < e

        cls.assert_that(
            lambda a, e: _less(a, e),
            error=error, that=f"'{actual}' is less than '{expected}'",
            actual=actual, expected=expected
        )

    @classmethod
    def assert_allin(cls, actual, expected, error=None):
        def _allin(a, e):
            return all(i in a for i in e)

        cls.assert_that(
            lambda a, e: _allin(a, e),
            error=error, that=f"'{actual}' are all in '{expected}'",
            actual=actual, expected=expected
        )

    @classmethod
    def assert_type(cls, actual, expected, error=None):
        cls.assert_that(
            lambda a, e: type(a) is e,
            error=error, that=f"type of '{actual}' is '{expected}'",
            actual=actual, expected=expected
        )


class JSONValidatorMixin(BaseAssert, JSONSchema):
    @classmethod
    def assert_schema(cls, data, schema):
        """

        :param data:
        :param schema:
        :return:
        """
        try:
            cls.validate(data, schema, raise_exc=True)
            valid, message = True, None
        except jsonschema.ValidationError as exc:
            valid, message = True, cls.error_report(exc, data)

        cls.assert_that(
            lambda a, e: a is True,
            actual=valid, error=f"Test that json is valid failed, got: {message}"
        )


class RegexMixin(BaseAssert):
    @classmethod
    def regex_find(cls, data, pattern, index=None):
        """

        :param data:
        :param pattern:
        :param index:
        :return:
        """
        occ = re.findall(pattern, data)
        if index is not None:
            if len(occ) > index:
                return occ[index]
            return None
        return occ

    @classmethod
    def regex_match(cls, data, pattern):
        """

        :param data:
        :param pattern:
        :return:
        """
        return bool(re.search(pattern, data))

    @classmethod
    def assert_match(cls, actual, expected, error=None):
        """

        :param actual:
        :param expected:
        :param error:
        """
        cls.assert_that(
            lambda a, e: cls.regex_match(a, e),
            error=error, that=f"'{actual}' matches '{expected}'",
            actual=actual, expected=expected
        )

    @classmethod
    def assert_occurrence(cls, actual, expected, occurrence,
                          error=None, greater=False, less=False):
        """

        :param actual:
        :param expected:
        :param occurrence:
        :param error:
        :param greater:
        :param less:
        """

        def find_all(a, e):
            tmp = len(cls.regex_find(a, e))
            if greater:
                return tmp > occurrence
            elif less:
                return tmp < occurrence
            else:
                return tmp == occurrence

        operator = "equals to"
        if greater:
            operator = "greater than"
        elif less:
            operator = "less than"

        that = f"occurrences of '{expected}' in '{actual}' are {operator} '{occurrence}'"
        cls.assert_that(
            lambda a, e: find_all(a, e),
            error=error, that=that, actual=actual, expected=expected
        )


class HttpAsserter(RegexMixin):
    @classmethod
    def assert_status_code(cls, response, code=200,
                           in_range=False, is_in=False, greater=False, less=False):
        """

        :param response:
        :param code:
        :param in_range:
        :param is_in:
        :param greater:
        :param less
        """
        status_code = response.status_code or response.status
        if type(code) in (list, tuple):
            if in_range is True:
                cls.assert_range(status_code, code)
            elif is_in is True:
                cls.assert_in(status_code, code)
            else:
                mess = "one of (is_in, in_range) must be true if a list of code is given"
                cls.assert_true(False, error=mess)
        else:
            if greater is True:
                cls.assert_greater(status_code, code)
            elif less is True:
                cls.assert_less(status_code, code)
            else:
                cls.assert_equals(status_code, code)

    @classmethod
    def assert_header(cls, response, name, value=None, is_in=False, regex=None):
        """

        :param response:
        :param name:
        :param value:
        :param is_in:
        :param regex:
        """
        header = response.headers.get(name)
        if is_in is True and value is not None:
            cls.assert_in(value, header)
        elif regex is not None:
            cls.assert_match(value, header)
        else:
            if is_in is True:
                cls.assert_in(name, response.headers)
            else:
                cls.assert_equals(value, header)

    def assert_headers(self, response, headers):
        """

        :param response:
        :param headers: {"name": {"value":"", "is_in": False, "regex": None}}
        :return:
        """
        for k, v in headers.items():
            self.assert_header(response, name=k, **v)

    @classmethod
    def assert_content_type(cls, response, value=None, is_in=True, regex=None):
        cls.assert_header(
            response, name='Content-Type',
            value=value, is_in=is_in, regex=regex
        )


class Asserter(
    HttpAsserter,
    JSONValidatorMixin,
):
    """single interface that inherits all Asserter mixins"""



import asyncio

import aiohttp
import flask

from flaskel import cap
from flaskel.utils.batch import AsyncBatchExecutor
from flaskel.utils.datastruct import ObjectDict
from flaskel.utils.uuid import get_uuid
from .client import HTTPBase, httpcode
from .httpdumper import FlaskelHTTPDumper


class HTTPBatch(HTTPBase, AsyncBatchExecutor):
    def __init__(self, conn_timeout=10, read_timeout=10, **kwargs):
        """

        :param conn_timeout:
        :param read_timeout:
        :param kwargs:
        """
        HTTPBase.__init__(self, **kwargs)
        AsyncBatchExecutor.__init__(self, return_exceptions=not self._raise_on_exc)
        self._timeout = aiohttp.ClientTimeout(sock_read=read_timeout, sock_connect=conn_timeout)

    async def http_request(self, dump_body=None, timeout=None, **kwargs):
        """

        :param dump_body:
        :param timeout:
        :param kwargs:
        :return:
        """
        if dump_body is None:
            dump_body = self._dump_body

        if timeout is None:
            timeout = self._timeout
        elif not isinstance(timeout, aiohttp.ClientTimeout):
            timeout = aiohttp.ClientTimeout(sock_read=timeout, sock_connect=timeout)

        try:
            self._logger.info(self.dump_request(ObjectDict(**kwargs), dump_body))
            async with aiohttp.ClientSession(timeout=timeout) as session, \
                    session.request(**kwargs) as resp:
                try:
                    body = await resp.json()
                except (aiohttp.ContentTypeError, ValueError, TypeError):
                    body = await resp.text()

                try:
                    response = ObjectDict(
                        body=body, status=resp.status,
                        headers={k: v for k, v in resp.headers.items()}
                    )
                    log_resp = response
                    log_resp.text = response.body
                    log_resp = self.dump_response(log_resp, dump_body)
                    resp.raise_for_status()
                    self._logger.info(log_resp)
                except aiohttp.ClientResponseError as exc:
                    self._logger.warning(log_resp)
                    if self._raise_on_exc is True:
                        raise  # pragma: no cover
                    response.exception = exc
                return response
        except (aiohttp.ClientError, aiohttp.ServerTimeoutError, asyncio.TimeoutError) as exc:
            self._logger.exception(exc)
            if self._raise_on_exc is True:
                raise  # pragma: no cover

            return ObjectDict(
                body={}, status=httpcode.SERVICE_UNAVAILABLE,
                headers={}, exception=exc
            )

    def request(self, requests, **kwargs):
        """

        :param requests:
        :return:
        """
        _requests = []
        for r in requests:
            r.setdefault('method', 'GET')
            self._tasks.append((self.http_request, r))

        return self.run()


class FlaskelHTTPBatch(HTTPBatch, FlaskelHTTPDumper):
    def __init__(self, **kwargs):
        kwargs.setdefault('logger', cap.logger)
        kwargs.setdefault('conn_timeout', cap.config.HTTP_TIMEOUT or 10)
        kwargs.setdefault('read_timeout', cap.config.HTTP_TIMEOUT or 10)
        super().__init__(**kwargs)

    def request(self, requests, **kwargs):
        if flask.request.id:
            for r in requests:
                if not r.get('headers'):
                    r['headers'] = {}
                req_id = f"{flask.request.id},{get_uuid()}"
                r['headers'][cap.config.REQUEST_ID_HEADER] = req_id

        return super().request(requests, **kwargs)



import io
from functools import partial

import flask
import jsonschema

from flaskel import cap


class Schemas:
    def __getattr__(self, item):
        return cap.config.SCHEMAS.get(item)


SCHEMAS = Schemas()


class JSONSchema:
    loader = partial(flask.json.loads)
    dumper = partial(flask.json.dumps, indent=4)
    marker = "3fb539deef7c4e2991f265c0a982f5ea"
    message_format = "{message}\nError in line {line}:\n{report}\n{message}"

    @classmethod
    def load_from_url(cls, url):
        """

        :param url:
        :return:
        """
        from flaskel.http import HTTPClient
        res = HTTPClient(url, raise_on_exc=True).get(url)
        return res.body

    @classmethod
    def load_from_file(cls, file):
        """

        :param file:
        :return:
        """
        with open(file[7:]) as f:
            return cls.loader(f.read())

    @classmethod
    def validate(cls, data, schema, raise_exc=False, pretty_error=True, checker=None):
        """

        :param data:
        :param schema:
        :param raise_exc:
        :param pretty_error:
        :param checker:
        :return:
        """
        if not schema:
            return True

        if type(schema) is str:
            if schema.startswith('http'):
                schema = cls.load_from_url(schema)
            if schema.startswith('file://'):
                schema = cls.load_from_file(schema)

        try:
            checker = checker or jsonschema.FormatChecker()
            jsonschema.validate(data, schema, format_checker=checker)
        except jsonschema.ValidationError as exc:
            if not raise_exc:
                if pretty_error:
                    cap.logger.error(cls.error_report(exc, data))
                else:
                    cap.logger.exception(exc)
                return False
            raise
        return True

    @classmethod
    def error_report(cls, e, json_object, lines_before=8, lines_after=8):
        """
        From: https://github.com/ccpgames/jsonschema-errorprinter/blob/master/jsonschemaerror.py

        Generate a detailed report of a schema validation error.
        'e' is a jsonschema.ValidationError exception that errored on
        'json_object'.
        Steps to discover the location of the validation error:
        1. Traverse the json object using the 'path' in the validation exception
           and replace the offending value with a special marker.
        2. Pretty-print the json object indendented json text.
        3. Search for the special marker in the json text to find the actual
           line number of the error.
        4. Make a report by showing the error line with a context of
          'lines_before' and 'lines_after' number of lines on each side.
        """
        if not e.path:
            return e.message or str(e)

        # Find the object that is erroring, and replace it with the marker.
        for entry in list(e.path)[:-1]:
            json_object = json_object[entry]

        orig, json_object[e.path[-1]] = json_object[e.path[-1]], cls.marker

        # Pretty print the object and search for the marker.
        json_error = cls.dumper(json_object)
        err_line = None

        for lineno, text in enumerate(io.StringIO(json_error)):
            if cls.marker in text:
                err_line = lineno
                break

        if not err_line:
            return e.message or str(e)

        report = []
        json_object[e.path[-1]] = orig
        json_error = cls.dumper(json_object)

        for lineno, text in enumerate(io.StringIO(json_error)):
            line_text = "{:4}: {}".format(lineno + 1, '>' * 3 if lineno == err_line else ' ' * 3)
            report.append(line_text + text.rstrip("\n"))

        report = report[max(0, err_line - lines_before):err_line + 1 + lines_after]
        return cls.message_format.format(
            line=err_line + 1,
            report="\n".join(report),
            message=e.message or str(e)
        )



from .api import api
from .auth import auth
from .test import test
from .web import web

BLUEPRINTS = (
    (api,),
    (test,),
    (web, {
        'url_prefix': '/'
    }),
    (auth, {
        'url_prefix': '/auth'
    }),
)



import os

import click
import decouple

from . import AppBuilder
from .utils import misc, yaml
from .utils.datastruct import ObjectDict
from .wsgi import BaseApplication, WSGIFactory


# noinspection PyUnusedLocal
def option_as_dict(ctx, param, value):
    """

    :param ctx:
    :param param:
    :param value:
    :return:
    """
    ret = {}
    for opt in value:
        k, v = opt.split("=", 2)
        ret.update({k: misc.parse_value(v)})
    return ret


class Server:
    opt_config_attr = dict(
        default=None,
        help='app yaml/json configuration file'
    )
    opt_log_config_attr = dict(
        default=None,
        help='alternative log yaml/json configuration file'
    )
    opt_bing_attr = dict(
        default='127.0.0.1:5000',
        help='address to bind',
        show_default=True
    )
    opt_debug_attr = dict(
        is_flag=True,
        flag_value=True,
        default=False,
        help='enable debug mode'
    )
    opt_wsgi_server_attr = dict(
        default=None,
        type=click.Choice(
            WSGIFactory.WSGI_SERVERS.keys(),
            case_sensitive=False
        ),
        help='name of wsgi server to use'
    )
    option_wsgi_config_attr = dict(
        default={},
        multiple=True,
        callback=option_as_dict,
        metavar="KEY=VAL",
        help='wsgi configuration'
    )
    option_app_config_attr = dict(
        default={},
        multiple=True,
        callback=option_as_dict,
        metavar="KEY=VAL",
        help='app configuration'
    )

    def __init__(self, app_factory=AppBuilder(), wsgi_factory=WSGIFactory()):
        """

        :param app_factory:
        :param wsgi_factory:
        """
        self._app_factory = app_factory
        self._wsgi_factory = wsgi_factory

        assert isinstance(self._app_factory, AppBuilder)
        assert isinstance(self._wsgi_factory, WSGIFactory)

    def _register_options(self, func):
        @click.command()
        @click.option('-b', '--bind', **self.opt_bing_attr)
        @click.option('-d', '--debug', **self.opt_debug_attr)
        @click.option('-c', '--config', **self.opt_config_attr)
        @click.option('-l', '--log-config', **self.opt_log_config_attr)
        @click.option('-w', '--wsgi-server', **self.opt_wsgi_server_attr)
        @click.option('-D', '--wsgi-config', **self.option_wsgi_config_attr)
        @click.option('-W', '--app-config', **self.option_app_config_attr)
        def _options(*args, **kwargs):
            return func(*args, **kwargs)

        return _options

    def run_from_cli(self, **kwargs):
        @self._register_options
        def forever(**params):
            self.serve_forever(**params, **kwargs)

        forever()

    @staticmethod
    def _prepare_config(filename, debug=None, bind=None, log_config=None):
        """

        :param filename:
        :param debug:
        :param bind:
        :param log_config:
        :return:
        """
        default_env = 'development' if debug else 'production'
        env = decouple.config('FLASK_ENV', default=default_env)

        if filename is not None:
            yaml.setup_yaml_parser()
            config = yaml.load_yaml_file(filename)
            if not config.app.FLASK_ENV:
                config.app.FLASK_ENV = env
        else:
            config = ObjectDict(
                app={'DEBUG': debug, 'FLASK_ENV': env},
                wsgi={'bind': bind, 'debug': debug}
            )

        if log_config is not None:
            config.app.LOG_FILE_CONF = log_config

        # debug flag enabled overrides config file
        if debug is True:
            config.app.DEBUG = True

        os.environ['FLASK_ENV'] = config.app.FLASK_ENV
        return config

    def serve_forever(self, config=None, log_config=None, bind=None, debug=None,
                      wsgi_class=None, wsgi_server=None, wsgi_config=None, app_config=None):
        """

        :param config: app and wsgi configuration file
        :param log_config: log configuration file
        :param bind: address to bind
        :param debug: enable debug mode
        :param wsgi_class: optional a custom subclass of BaseApplication
        :param wsgi_server: wsgi server chosen
        :param wsgi_config: wsgi configuration
        :param app_config: app configuration
        :return: never returns
        """
        config = self._prepare_config(config, debug, bind, log_config)
        wsgi_config = {**(config.get('wsgi') or {}), **(wsgi_config or {})}
        app_config = {**(config.get('app') or {}), **(app_config or {})}

        if wsgi_server:
            wsgi_class = self._wsgi_factory.get_class(wsgi_server)
        elif wsgi_class is None:
            wsgi_class = self._wsgi_factory.get_class('builtin')

        assert issubclass(wsgi_class, BaseApplication)

        app = self._app_factory.get_or_create(app_config)
        wsgi = wsgi_class(app, options=wsgi_config)
        wsgi.run()  # run forever



from gunicorn.app.base import BaseApplication as WSGIServer

from .base import BaseApplication


class WSGIGunicorn(BaseApplication, WSGIServer):
    def __init__(self, app, options=None):
        """

        :param app:
        :param options:
        """
        BaseApplication.__init__(self, app, options)
        WSGIServer.__init__(self)

    def load_config(self):
        """

        """
        options = {}

        for k, v in self.options.items():
            if k in self.cfg.settings and v is not None:
                options.update({k: v})

        for key, value in options.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        """

        :return:
        """
        return self.application

    def init(self, parser, opts, args):
        """

        :param parser:
        :param opts:
        :param args:
        """
        # abstract in superclass

    def run(self):
        """

        """
        WSGIServer.run(self)



# Define your extensions in this package and import them here
from flaskel.ext import caching, cors, ip_ban, limit
from flaskel.ext.sqlalchemy import db

EXTENSIONS = {
    # "name":   (<extension>, parameters: dict)
    "cors":     (cors,),
    "database": (db,),
    "template": None,
    "limiter":  (limit.limiter,),
    "ip_ban":   (ip_ban,),
    "cache":    (caching,),
}



import socket

from flask_mail import Mail, Message


class ClientMail(Mail):
    def sendmail(self, app, sender=None, recipients=None, message=None, attachments=None, **kwargs):
        """

        :param app:
        :param message:
        :param sender:
        :param recipients:
        :param attachments:
        """
        destination = recipients or [app.config.MAIL_RECIPIENT]
        if sender:
            try:
                sender = (f"{sender.name} {sender.surname}", sender.email)
            except AttributeError:
                pass
        else:
            sender = app.config.MAIL_DEFAULT_SENDER

        mail_message = Message(sender=sender, recipients=destination, **kwargs)
        mail_message.html = message

        if attachments:
            try:
                for attach in attachments:
                    with open(attach['filename']) as f:
                        mail_message.attach(data=f.read(), **(attach or {}))
            except (OSError, IOError) as exc:
                app.logger.warning(str(exc))

        try:
            socket.setdefaulttimeout(app.config.MAIL_TIMEOUT or 60)
            with app.app_context():
                self.send(mail_message)
        except OSError as exc:
            app.logger.exception(exc)

        app.logger.info(f"email {mail_message.msgId} from {sender} sent to {destination}")


client_mail = ClientMail()



__version__ = '0.0.0'

__author_name__ = 'author name'
__author_email__ = 'author@email.com'
__url_version__ = '/v' + __version__.split('.')[0]

__author__ = f"{__author_name__} <{__author_email__}>"

__all__ = [
    '__version__',
    '__author__',
    '__author_name__',
    '__author_email__',
    '__url_version__',
]



from flask_sqlalchemy import Model


class SQLAModel(Model):
    @classmethod
    def get_one(cls, raise_not_found=True, to_dict=True, *args, **kwargs):
        """

        :param raise_not_found:
        :param to_dict:
        :return:
        """
        res = cls.query.filter(*args).filter_by(**kwargs)

        if raise_not_found:
            res = res.first_or_404()
        else:
            res = res.first()
            if res is None:
                return

        if to_dict is True:
            return res.to_dict()
        else:
            return res

    @classmethod
    def get_list(cls, to_dict=True, order_by=None,
                 page=None, page_size=None, max_per_page=None, *args, **kwargs):
        """

        :param to_dict:
        :param order_by:
        :param page:
        :param page_size:
        :param max_per_page:
        :return:
        """
        q = cls.query.filter(*args).filter_by(**kwargs)

        if order_by is not None:
            q = q.order_by(order_by)

        if page or page_size:
            q = q.paginate(page, page_size, False, max_per_page)
            res = q.items
            if to_dict is False:
                return q
        else:
            res = q.all()

        if to_dict is True:
            return [r.to_dict() for r in res]
        return res



import re
from datetime import datetime

import flask
from flask_limiter import Limiter

from flaskel import cap, httpcode
from flaskel.ext import cfremote
from flaskel.utils.datastruct import ObjectDict
from flaskel.utils.datetime import Day


def response_ok(r):
    return httpcode.is_ok(r.status_code)


def response_ko(r):
    return r.status_code != httpcode.TOO_MANY_REQUESTS \
           and httpcode.is_ko(r.status_code)


limiter = Limiter(
    key_func=cfremote.get_remote,
    default_limits=[lambda: cap.config.LIMITER.FAIL],
    default_limits_deduct_when=response_ko
)


def header_whitelist():
    hdr = cap.config.LIMITER.BYPASS_KEY
    value = cap.config.LIMITER.BYPASS_VALUE
    return flask.request.headers.get(hdr) == value


@limiter.request_filter
def _header_whitelist():
    return header_whitelist()


class RateLimit:  # pragma: no cover
    limiter = limiter

    @classmethod
    def slow(cls):
        return cls.limiter.limit(
            lambda: cap.config.LIMITER.SLOW,
            deduct_when=response_ok
        )

    @classmethod
    def medium(cls):
        return cls.limiter.limit(
            lambda: cap.config.LIMITER.MEDIUM,
            deduct_when=response_ok
        )

    @classmethod
    def fast(cls):
        return cls.limiter.limit(
            lambda: cap.config.LIMITER.FAST,
            deduct_when=response_ok
        )

    @classmethod
    def fail(cls):
        return cls.limiter.limit(
            lambda: cap.config.LIMITER.FAIL,
            deduct_when=response_ko
        )


class FlaskIPBan:
    def __init__(self, app=None, **kwargs):
        self._ip_banned = {}
        self._url_blocked = {}

        self._ip_whitelist = {
            '127.0.0.1': True
        }

        self._url_whitelist = {
            '^/.well-known/': ObjectDict(pattern=re.compile(r'^/.well-known'), match_type='regex'),
            '/favicon.ico':   ObjectDict(pattern=re.compile(''), match_type='string'),
            '/robots.txt':    ObjectDict(pattern=re.compile(''), match_type='string'),
            '/ads.txt':       ObjectDict(pattern=re.compile(''), match_type='string'),
        }

        if app:
            self.init_app(app, **kwargs)  # pragma: no cover

    def init_app(self, app, whitelist=(), nuisances=None):
        """

        :param app: flask app
        :param whitelist:
        :param nuisances:
        :return:
        """
        app.config.setdefault('IPBAN_COUNT', 20)
        app.config.setdefault('IPBAN_SECONDS', Day.seconds)
        app.config.setdefault('IPBAN_NUISANCES', nuisances or {})
        app.config.setdefault('IPBAN_STATUS_CODE', httpcode.FORBIDDEN)
        app.config.setdefault('IPBAN_CHECK_CODES', (
            httpcode.NOT_FOUND, httpcode.METHOD_NOT_ALLOWED, httpcode.NOT_IMPLEMENTED
        ))

        app.after_request(self._after_request)
        app.before_request(self._before_request)

        self.add_whitelist(whitelist or [])
        self.load_nuisances(conf=cap.config.IPBAN_NUISANCES)

        if not hasattr(app, 'extensions'):
            app.extensions = dict()  # pragma: no cover
        app.extensions['ipban'] = self

    @staticmethod
    def get_ip():
        return flask.request.remote_addr

    @staticmethod
    def get_url():
        return flask.request.path

    def _is_excluded(self, ip=None, url=None):
        """

        :return: true if this ip or url should not be checked
        """
        ip = ip or self.get_ip()
        url = url or self.get_url()

        for key, item in self._url_whitelist.items():
            if (item.match_type == 'regex' and item.pattern.match(url)) \
                    or (item.match_type == 'string' and key == url):
                return True

        if ip in self._ip_whitelist:
            return True

        return False

    def _test_blocked(self, url='', ip=None):
        """
        return true if the url or ip pattern matches an existing block

        :param url: (optional) the url to check
        :param ip: (optional) an ip to check
        :return:
        """
        query_path = url.split('?')[0]
        for pattern, item in self._url_blocked.items():
            if item.match_type == 'regex' and item.pattern.match(query_path):
                cap.logger.warning(f'url {url} matches block pattern {pattern}')
                return True
            elif item.match_type == 'string' and pattern == query_path:
                cap.logger.warning(f'url {url} matches block string {pattern}')
                return True
            elif ip and item.match_type == 'ip' and pattern == ip:
                cap.logger.warning(f'ip block match {ip}')
                return True

        return False

    def _after_request(self, response):
        """
        method to call after a request to allow recording all unwanted status codes

        :param response:
        :return:
        """
        if response.status_code in cap.config.IPBAN_CHECK_CODES:
            self.add()

        return response

    def _before_request(self):
        if self._is_excluded():
            return

        ip = self.get_ip()
        url = self.get_url()
        entry = self._ip_banned.get(ip)

        if entry and (entry.count or 0) > cap.config.IPBAN_COUNT:
            if entry.permanent:
                flask.abort(cap.config.IPBAN_STATUS_CODE)

            now = datetime.now()
            delta = now - (entry.timestamp or now)
            if delta.seconds < cap.config.IPBAN_SECONDS or cap.config.IPBAN_SECONDS == 0:
                cap.logger.info(f"IP: {ip} updated in ban list, at url: {url}")
                entry.count += 1
                entry.timestamp = now
                flask.abort(cap.config.IPBAN_STATUS_CODE)

            entry.count = 0
            cap.logger.debug(f"IP: {ip} expired from ban list, at url: {url}")

    def add_url_block(self, url, match_type='regex'):
        """
        add or replace the pattern to the list of url patterns to block

        :param match_type: regex or string - determines the match strategy to use
        :param url: regex pattern to match with requested url
        :return: length of the blocked list
        """
        self._url_blocked[url] = ObjectDict(pattern=re.compile(url), match_type=match_type)
        return len(self._url_blocked)

    def block(self, ips, permanent=False, timestamp=None, url='block'):
        """
        add a list of ip addresses to the block list

        :param ips: list of ip addresses to block
        :param permanent: (optional) True=do not allow entries to expire
        :param timestamp: use this timestamp instead of now()
        :param url: url or reason to block
        :returns number of entries in the block list
        """
        timestamp = timestamp or datetime.now()

        for ip in ips:
            entry = self._ip_banned.get(ip)
            if entry:
                entry.timestamp = timestamp
                entry.count = cap.config.IPBAN_COUNT * 2
                # retain permanent on extra blocks
                entry.permanent = entry.permanent or permanent
                cap.logger.warning(f'{ip} added to ban list')
            else:
                self._ip_banned[ip] = ObjectDict(
                    timestamp=timestamp, count=cap.config.IPBAN_COUNT * 2,
                    permanent=permanent, url=url
                )
                cap.logger.info(f'{ip} updated in ban list')

        return len(self._ip_banned)

    def add(self, ip=None, url=None, timestamp=None):
        """
        increment ban count ip of the current request in the banned list

        :return:
        :param ip: optional ip to add (ip ban will by default use current ip)
        :param url: optional url to display/store
        :param timestamp: entry time to set
        :return True if entry added/updated
        """
        ip = ip or self.get_ip()
        url = url or self.get_url()

        if self._is_excluded(ip=ip, url=url):
            return False

        entry = self._ip_banned.get(ip)
        # check url block list if no existing entry or existing entry has expired
        if not entry or (entry and (entry.count or 0) < cap.config.IPBAN_COUNT):
            if self._test_blocked(url, ip=ip):
                self.block([ip], url=url)
                return True

        if not timestamp or (timestamp and timestamp > datetime.now()):
            timestamp = datetime.now()

        if entry:
            entry.timestamp = timestamp
            count = entry.count = entry.count + 1
        else:
            count = 1
            self._ip_banned[ip] = ObjectDict(timestamp=timestamp, count=count, url=url)

        cap.logger.info(f"{ip} {url} added/updated ban list. Count: {count}")
        return True

    def remove(self, ip):
        """
        remove from the ban list

        :param ip: ip to remove
        :return True if entry removed
        """
        if not self._ip_banned.get(ip):
            return False

        del self._ip_banned[ip]
        return True

    def load_nuisances(self, conf=None):
        """
        load a yaml file of nuisance urls or ips that are commonly used by vulnerability scanners
        any access to one of these that produces unwanted status codes will ban the source ip.
        Each call to load_nuisances will add to the current list of nuisances

        :param conf:
        :return: the number of nuisances added from this file
        """
        count = 0
        for match_type in ['ip', 'string', 'regex']:
            values = conf.get(match_type) or ()
            for v in values:
                try:
                    self.add_url_block(v, match_type)
                    count += 1
                except Exception as e:
                    cap.logger.exception(f"Exception {e} adding pattern {v}")

        return count

    def add_whitelist(self, ips):
        """

        :param ips: list of ip addresses to add
        :return: number of entries in the ip whitelist
        """
        if not isinstance(ips, (list, tuple)):
            ips = (ips,)

        for ip in ips:
            self._ip_whitelist[ip] = True

        return len(self._ip_whitelist)

    def remove_whitelist(self, el):
        if not self._ip_whitelist.get(el):
            return False

        del self._ip_whitelist[el]
        return True



import functools
import inspect

import flask
from flask.views import View

import flaskel.http.rpc as rpc
from flaskel import cap, Response, httpcode
from flaskel.ext import builder
from flaskel.utils.batch import DaemonThread, ThreadBatchExecutor
from flaskel.utils.datastruct import ObjectDict


class JSONRPCView(View):
    version = '2.0'
    separator = '.'
    methods = ['POST']
    decorators = [builder.response('json')]

    def __init__(self, operations=None, batch_executor=None, **kwargs):
        """

        :param batch_executor
        :param operations: a dict of jsonrpc operations like:
            {
                None: {
                    "action1": funct1,
                    "action2": funct2,
                    ...
                },
                "method1": {
                    "action1": funct1,
                    "action2": funct2,
                    ...
                },
                ...
            }
        """
        self.operations = operations or {}
        self._batch_executor = batch_executor or ThreadBatchExecutor
        kwargs.setdefault('thread_class', DaemonThread)
        self._batch_args = kwargs

    def _validate_request(self, data):
        if 'jsonrpc' not in data or 'method' not in data:
            raise rpc.RPCInvalidRequest() from None

        if data['jsonrpc'] != self.version:
            raise rpc.RPCInvalidRequest(
                f"jsonrpc version is {self.version}",
                req_id=data.get('id')
            ) from None

    def _validate_payload(self):
        payload = flask.request.get_json(True)
        if not payload:
            raise rpc.RPCParseError() from None

        if type(payload) is list:
            max_requests = cap.config.JSONRPC_BATCH_MAX_REQUEST
            if max_requests and len(payload) > max_requests:
                mess = f"Operations in a single http request must be less than {max_requests}"
                flask.abort(
                    httpcode.REQUEST_ENTITY_TOO_LARGE,
                    response=dict(reason=mess)
                )
            for d in payload:
                self._validate_request(d)
        else:
            self._validate_request(payload)

        return payload

    def _get_action(self, method):
        """

        :param method:
        :return:
        """
        m = method.split(self.separator)
        if len(m) > 1:
            op, action = m[0], self.separator.join(m[1:])
        else:
            op, action = None, m[0]

        try:
            return self.operations[op][action]
        except (IndexError, TypeError, KeyError) as exc:
            cap.logger.debug(exc)
            raise rpc.RPCMethodNotFound()

    def dispatch_request(self):
        """

        :return:
        """
        tasks = []
        responses = []

        try:
            payload = self._validate_payload()
        except rpc.RPCError as ex:
            return ObjectDict(
                jsonrpc=self.version,
                id=getattr(ex, 'req_id', None),
                error=ex.as_dict()
            ), httpcode.BAD_REQUEST

        for d in (payload if type(payload) is list else [payload]):
            resp = ObjectDict(jsonrpc=self.version, id=None)
            try:
                if 'id' not in d:
                    tasks.append((self._get_action(d['method']), {**(d.get('params') or {})}))
                else:
                    resp.id = d.get('id')
                    action = self._get_action(d['method'])
                    resp.result = action(**(d.get('params') or {}))
            except rpc.RPCError as ex:
                resp.error = ex.as_dict()
            except Exception as ex:
                cap.logger.exception(ex)
                mess = str(ex) if cap.debug is True else None
                resp.error = rpc.RPCInternalError(message=mess).as_dict()

            if 'id' in d:
                responses.append(resp)

        self._batch_executor(tasks=tasks, **self._batch_args).run()

        if not responses:
            res = Response.no_content()
            return None, res.status_code, res.headers

        if type(payload) is list:
            if len(responses) > 1:
                return responses, httpcode.MULTI_STATUS
            return responses

        return responses[0]

    def load_from_object(self, obj):
        """

        :param obj:
        """
        for m in inspect.getmembers(obj, predicate=inspect.isroutine):
            if not m[0].startswith('_'):
                self.method(obj.__class__.__name__, m[0])(m[1])

    def method(self, name=None, operation=None):
        """

        :param name:
        :param operation:
        :return:
        """

        def _method(func):
            nonlocal name
            name = name or func.__name__

            @functools.wraps(func)
            def wrapped():
                obj = {operation: func}
                if name not in self.operations:
                    self.operations[name] = obj
                else:
                    self.operations[name].update(obj)

            return wrapped()

        return _method

    def register(self, app, name=None, url=None, **kwargs):
        """

        :param app: Flask or Blueprint instance
        :param name: view name
        :param url: url to bind
        """
        name = name or self.__class__.__name__
        url = url or f"/{name}"
        view_func = self.__class__.as_view(name, self.operations, **kwargs)
        app.add_url_rule(f"/{url.rstrip('/')}", view_func=view_func, methods=self.methods)



import flask

from flaskel.http import UserAgent as UserAgentParser


class UserAgent:
    def __init__(self, app=None, **kwargs):
        """

        :param app:
        :param kwargs:
        """
        self._parser = None

        if app is not None:
            self.init_app(app, **kwargs)  # pragma: no cover

    def init_app(self, app, parser_class=UserAgentParser):
        """

        :param app:
        :param parser_class:
        """
        self._parser = parser_class
        app.config.setdefault('USER_AGENT_AUTO_PARSE', False)

        if not hasattr(app, 'extensions'):
            app.extensions = {}  # pragma: no cover
        app.extensions['useragent'] = self

        @app.before_request
        def before_request():
            flask.g.user_agent = self._parser(flask.request.user_agent.string)
            if app.config['USER_AGENT_AUTO_PARSE']:
                flask.g.user_agent.parse()


useragent = UserAgent()



import flask
from sqlalchemy.exc import IntegrityError

from flaskel import cap, httpcode
from flaskel.utils import webargs
from .base import Resource


class CatalogResource(Resource):
    methods_collection = ['GET', 'HEAD']
    methods_resource = ['GET', 'HEAD']
    methods_subresource = ['GET', 'HEAD']

    def __init__(self, model):
        self._model = model

    def on_get(self, res_id, *args, **kwargs):
        return self._model.get_one(id=res_id)

    def on_collection(self, params=None, model=None, *args, **kwargs):
        if params is None:
            params = webargs.paginate()

        model = model or self._model
        max_size = cap.config.MAX_PAGE_SIZE
        page = params.get('page')
        size = params.get('page_size') or max_size
        order_by = getattr(model, 'order_by', None)

        return self.response_paginated(
            model.get_list(
                to_dict=False, order_by=order_by, page=page,
                page_size=size, max_per_page=max_size, **kwargs
            ),
            restricted=not params.get('related', False)
        )

    def response_paginated(self, res, **kwargs):
        if type(res) is list:
            return [r.to_dict(**kwargs) for r in res]

        return (
            [r.to_dict(**kwargs) for r in res.items],
            httpcode.PARTIAL_CONTENT if res.has_next else httpcode.SUCCESS,
            self.pagination_headers(res)
        )

    @staticmethod
    def pagination_headers(data):
        return {
            'X-Pagination-Count':     data.total,
            'X-Pagination-Page':      data.page,
            'X-Pagination-Num-Pages': data.pages,
            'X-Pagination-Page-Size': data.per_page,
        }


class Restful(CatalogResource):
    """TODO Not tested yet"""
    methods_collection = ['POST', 'GET', 'HEAD']
    methods_subresource = ['POST', 'GET', 'HEAD']
    methods_resource = ['GET', 'HEAD', 'PUT', 'DELETE']

    def __init__(self, db, model):
        """

        :param db:
        :param model:
        """
        super().__init__(model)
        self._db = db

    def on_post(self, payload=None, *args, **kwargs):
        """

        :param payload:
        :return:
        """
        res = self._model(**(payload or {}))
        try:
            self._db.session.add()
            self._db.session.commit()
        except IntegrityError:
            self._db.session.rollback()
            flask.abort(httpcode.CONFLICT)
        finally:
            self._db.session.close()

        return res.to_dict(), httpcode.CREATED

    def on_delete(self, res_id, *args, **kwargs):
        """

        :param res_id:
        """
        res = self._model.query.get_or_404(res_id)
        res.delete()
        self._db.session.commit()

    def on_put(self, res_id, payload=None, *args, **kwargs):
        """

        :param res_id:
        :param payload:
        :return:
        """
        res = self._model.query.get_or_404(res_id)
        res.update(**(payload or {}))
        self._db.session.add(res)
        self._db.session.commit()
        return res.to_dict()



from werkzeug.routing import BaseConverter, FloatConverter, UnicodeConverter


class ListConverter(BaseConverter):
    default_separator = '+'

    def __init__(self, map_name=None, sep=None):
        """

        :param map_name:
        :param sep:
        """
        super().__init__(map_name)
        self._sep = sep or self.default_separator

    def to_python(self, value):
        """

        :param value:
        :return:
        """
        return value.split(self._sep)

    def to_url(self, values):
        """

        :param values:
        :return:
        """
        return self._sep.join(
            BaseConverter.to_url(self, v) for v in values
        )


CONVERTERS = {
    'list': ListConverter,
    'str': UnicodeConverter,
    'decimal': FloatConverter,
}



class FakeLogger:
    def debug(self, *args, **kwargs):
        pass

    def info(self, *args, **kwargs):
        pass

    def warning(self, *args, **kwargs):
        pass

    def error(self, *args, **kwargs):
        pass

    def exception(self, *args, **kwargs):
        pass


# noinspection PyMethodMayBeStatic,PyUnusedLocal
class DummyLogger:
    def debug(self, *args, **kwargs):
        print("DEBUG:", *args)

    def info(self, *args, **kwargs):
        print("INFO:", *args)

    def warning(self, *args, **kwargs):
        print("WARNING:", *args)

    def error(self, *args, **kwargs):
        print("ERROR:", *args)

    def exception(self, exc, *args, **kwargs):
        print("CRITICAL:", *args)



from decouple import config

DEBUG = config('DEBUG', default=False, cast=bool)
TESTING = config('TESTING', default=DEBUG, cast=bool)
APP_NAME = config('APP_NAME', default="flaskel")
APP_HOST = config('APP_HOST', default='127.0.0.1')
APP_PORT = config('APP_PORT', default=5000, cast=int)
FLASK_APP = config('FLASK_APP', default="app:app")
FLASK_ENV = config('FLASK_ENV', default='development' if DEBUG else 'production')
SERVER_NAME = config('SERVER_NAME', default=f"{APP_HOST}:{APP_PORT}")

LOCALE = config('LOCALE', default="en_EN.utf8")
TEMPLATES_AUTO_RELOAD = config("TEMPLATES_AUTO_RELOAD", default=DEBUG, cast=bool)
EXPLAIN_TEMPLATE_LOADING = config("EXPLAIN_TEMPLATE_LOADING", default=DEBUG, cast=bool)

SECRET_KEY_MIN_LENGTH = 256

AC_ALLOW_ORIGIN = []

RB_DEFAULT_ACCEPTABLE_MIMETYPES = [
    "application/json",
    "application/xml",
]

ERROR_PAGE = 'core/error.html'
LOG_LOGGER_NAME = FLASK_ENV
LOG_APP_NAME = APP_NAME

CF_STRICT_ACCESS = False

USE_X_SENDFILE = DEBUG
SEND_FILE_MAX_AGE_DEFAULT = 86400
ENABLE_ACCEL = True
ACCEL_BUFFERING = True
ACCEL_CHARSET = 'utf-8'
ACCEL_LIMIT_RATE = 'off'

WSGI_WERKZEUG_LINT_ENABLED = DEBUG
WSGI_WERKZEUG_PROFILER_ENABLED = DEBUG
WSGI_WERKZEUG_PROFILER_RESTRICTION = (0.1,)
WSGI_WERKZEUG_PROFILER_FILE = 'profiler.txt'

DATE_ISO_FORMAT = "%Y-%m-%dT%H:%M:%S"

REQUEST_ID_HEADER = 'X-Request-ID'
RATELIMIT_HEADERS_ENABLED = True
LIMITER = {
    'FAIL':         '1/second',
    'FAST':         '30/minute',
    'MEDIUM':       '20/minute',
    'SLOW':         '10/minute',
    'BYPASS_KEY':   'X-Limiter-Bypass',
    'BYPASS_VALUE': 'bypass-rate-limit',
}



from flaskel.utils.datastruct import ObjectDict


class RPCError(Exception):
    def __init__(self, code, message, data=None):
        """

        :param code:
        :param message:
        :param data:
        """
        self.code = code
        self.message = message
        self.data = data

    def as_dict(self):
        """

        :return:
        """
        return ObjectDict(code=self.code, message=self.message, data=self.data)


class RPCParseError(RPCError):
    def __init__(self, message='Invalid JSON was received by the server', data=None):
        """

        :param message:
        :param data:
        """
        super().__init__(-32700, message, data)


class RPCInvalidRequest(RPCError):
    def __init__(self, message='The JSON sent is not a valid Request object', data=None, req_id=None):
        """

        :param message:
        :param data:
        """
        super().__init__(-32600, message, data)
        self.req_id = req_id


class RPCMethodNotFound(RPCError):
    def __init__(self, message='The method does not exist or is not available', data=None):
        """

        :param message:
        :param data:
        """
        super().__init__(-32601, message, data)


class RPCInvalidParams(RPCError):
    def __init__(self, message='Invalid method parameter(s)', data=None):
        """

        :param message:
        :param data:
        """
        super().__init__(-32602, message, data)


class RPCInternalError(RPCError):
    def __init__(self, message='Internal JSON-RPC error', data=None):
        """

        :param message:
        :param data:
        """
        super().__init__(-32603, message, data)



import enum

CONTINUE = 100
SWITCHING_PROTOCOL = 101
PROCESSING = 102  # WebDAV
EARLY_HINTS = 103
SUCCESS = 200
CREATED = 201
ACCEPTED = 202
NON_AUTHORITATIVE_INFORMATION = 203
NO_CONTENT = 204
RESET_CONTENT = 205
PARTIAL_CONTENT = 206
MULTI_STATUS = 207  # WebDAV
ALREADY_REPORTED = 208  # WebDAV
IM_USED = 226
MULTIPLE_CHOICES = 300
MOVED_PERMANENTLY = 301
FOUND = 302
SEE_OTHER = 303
NOT_MODIFIED = 304
USE_PROXY = 305
RESERVED = 306
TEMPORARY_REDIRECT = 307
PERMANENT_REDIRECT = 308
BAD_REQUEST = 400
UNAUTHORIZED = 401
PAYMENT_REQUIRED = 402
FORBIDDEN = 403
NOT_FOUND = 404
METHOD_NOT_ALLOWED = 405
NOT_ACCEPTABLE = 406
PROXY_AUTHENTICATION_REQUIRED = 407
REQUEST_TIMEOUT = 408
CONFLICT = 409
GONE = 410
LENGTH_REQUIRED = 411
PRECONDITION_FAILED = 412
REQUEST_ENTITY_TOO_LARGE = 413
REQUEST_URI_TOO_LONG = 414
UNSUPPORTED_MEDIA_TYPE = 415
RANGE_NOT_SATISFIABLE = 416
EXPECTATION_FAILED = 417
IM_A_TEAPOT = 418
MISDIRECTED_REQUEST = 421
UNPROCESSABLE_ENTITY = 422
LOCKED = 423
FAILED_DEPENDENCY = 424  # WebDAV
TOO_EARLY = 425
UPGRADE_REQUIRED = 426
PRECONDITION_REQUIRED = 428
TOO_MANY_REQUESTS = 429
REQUEST_HEADER_FIELDS_TOO_LARGE = 431
UNAVAILABLE_FOR_LEGAL_REASON = 451
INTERNAL_SERVER_ERROR = 500
NOT_IMPLEMENTED = 501
BAD_GATEWAY = 502
SERVICE_UNAVAILABLE = 503
GATEWAY_TIMEOUT = 504
HTTP_VERSION_NOT_SUPPORTED = 505
INSUFFICIENT_STORAGE = 507  # WebDAV
LOOP_DETECTED = 508  # WebDAV
NOT_EXTENDED = 510
NETWORK_AUTHENTICATION_REQUIRED = 511


class StatusType(enum.IntEnum):
    INFORMATIONAL = 1
    SUCCESS = 2
    REDIRECTION = 3
    CLIENT_ERROR = 4
    SERVER_ERROR = 5


def status_type(s):
    return int(s / 100)


def is_informational(s):
    return status_type(s) == StatusType.INFORMATIONAL


def is_success(s):
    return status_type(s) == StatusType.SUCCESS


def is_redirection(s):
    return status_type(s) == StatusType.REDIRECTION


def is_client_error(s):
    return status_type(s) == StatusType.CLIENT_ERROR


def is_server_error(s):
    return status_type(s) == StatusType.SERVER_ERROR


def is_ok(s):
    return status_type(s) in (
        StatusType.INFORMATIONAL,
        StatusType.SUCCESS,
        StatusType.REDIRECTION
    )


def is_ko(s):
    return not is_ok(s)



import os
import shutil
import sys
from pathlib import Path

import click

from .ext.sqlalchemy import db_to_schema, model_to_uml
from .tester import cli as cli_tester

INIT_CONTENT = """
# generated via cli
from .version import *
"""


@click.group()
def cli():
    """

    """
    pass


@cli.command(name='init')
@click.argument('name')
def init(name):
    """Create skeleton for new application"""
    import flaskel

    package_path = flaskel.__path__[0]
    source = os.path.join(package_path, 'skeleton')
    destination = name

    try:
        shutil.copytree(source, destination)

        init_file = Path(os.path.join(destination, '__init__.py'))
        init_file.write_text(INIT_CONTENT)

        if not os.path.isfile('setup.py'):
            shutil.move(os.path.join(destination, 'setup.py'), '.')
            setup_file = Path('setup.py')
            text = setup_file.read_text()
            text = text.replace('{skeleton}', name)
            setup_file.write_text(text)
        else:
            os.remove(os.path.join(destination, 'setup.py'))

        if not os.path.isdir('config'):
            shutil.move(os.path.join(destination, 'config'), '.')
        else:
            os.remove(os.path.join(destination, 'config'))

        if not os.path.isfile('Dockerfile'):
            shutil.move(os.path.join(destination, 'Dockerfile'), '.')
        else:
            os.remove(os.path.join(destination, 'Dockerfile'))

        if not os.path.isdir('tests'):
            shutil.move(os.path.join(destination, 'tests'), '.')
            cli_file = Path(os.path.join('tests', '__init__.py'))
            text = cli_file.read_text()
            text = text.replace('from ext', f"from {name}.ext")
            text = text.replace('from blueprint', f"from {name}.blueprint")
            cli_file.write_text(text)
        else:
            os.remove(os.path.join(destination, 'tests'))

        cli_file = Path(os.path.join(destination, 'cli.py'))
        text = cli_file.read_text()
        text = text.replace('from ext', f"from {name}.ext")
        text = text.replace('from blueprint', f"from {name}.blueprint")
        cli_file.write_text(text)
    except OSError as e:
        print(f"Unable to create new app. Error: {e}", file=sys.stderr)


@cli.command(name='tests')
def run_tests():
    """Run test suite from package tests in current directory"""
    cli_tester.main()


@cli.command(name='schema')
@click.option('-m', '--from-models', help='module of sqlalchemy models')
@click.option('-d', '--from-database', help='database url connection')
@click.option('-f', '--file', required=True, help='output png filename')
def dump_schema(from_models, from_database, file):
    """Create png schema of database"""
    if from_models:
        graph = model_to_uml(from_models)
    elif from_database:
        graph = db_to_schema(from_database)
    else:
        raise click.UsageError("One of -m or -d are required")

    graph.write_png(file)


if __name__ == '__main__':
    cli()



# based on:
# https://github.com/nameko/nameko/blob/a719cb1487f643769e2d13daf255c20551490f43/nameko/cli/main.py#L102-L111

import os
import re
from functools import partial

import yaml

from .datastruct import ObjectDict

ENV_VAR_MATCHER = re.compile(
    r"""
        \${        # match characters `${` literally
        ([^}:\s]+) # 1st group: matches any character except `}` or `:`
        :?         # matches the literal `:` character zero or one times
        ([^}]+)?   # 2nd group: matches any character except `}`
        }          # match character `}` literally
    """, re.VERBOSE
)

IMPLICIT_ENV_VAR_MATCHER = re.compile(
    r"""
        .*      # matches any number of any characters
        \${.*}  # matches any number of any characters
                # between `${` and `}` literally
        .*      # matches any number of any characters
    """, re.VERBOSE
)


def _replace_env_var(match):
    env_var, default = match.groups()
    value = os.environ.get(env_var, None)
    if value is None:
        # expand default using other vars
        if default is None:
            # regex module return None instead of
            #  '' if engine didn't entered default capture group
            default = ''

        value = default
        while IMPLICIT_ENV_VAR_MATCHER.match(value):
            value = ENV_VAR_MATCHER.sub(_replace_env_var, value)
    return value


def env_var_constructor(loader, node, raw=False):
    raw_value = loader.construct_scalar(node)
    value = ENV_VAR_MATCHER.sub(_replace_env_var, raw_value)
    return value if raw else yaml.safe_load(value)


def include_constructor(loader, node):
    with open(node.value) as f:
        return yaml.load(f, Loader=loader.__class__)


def setup_yaml_parser():
    yaml.add_constructor('!include', include_constructor)
    yaml.add_constructor('!env_var', env_var_constructor)
    yaml.add_constructor('!raw_env_var', partial(env_var_constructor, raw=True))
    yaml.add_implicit_resolver('!env_var', IMPLICIT_ENV_VAR_MATCHER)


def load_yaml_file(filename):
    with open(filename) as f:
        return ObjectDict.normalize(yaml.load(f, Loader=yaml.Loader))


def loads(data):
    return ObjectDict.normalize(yaml.load(data, Loader=yaml.Loader))



from functools import partial

from flask import request
from flask_caching import Cache as FlaskCache

from flaskel import httpcode

caching = FlaskCache()


class Cache:
    cache = caching

    cached = partial(
        cache.cached,
        unless=lambda: 'no-store' in (request.headers.get('Cache-Control') or ''),
        response_filter=lambda r: httpcode.is_ok(r.status_code)
    )



from datetime import timedelta

import flask_jwt_extended as jwt
from flask_httpauth import HTTPBasicAuth

from flaskel import cap, httpcode
from flaskel.utils.datastruct import ObjectDict
from .sqlalchemy import db

jwtm = jwt.JWTManager()
basic_auth = HTTPBasicAuth()


@basic_auth.verify_password
def simple_basic_auth(username, password):
    if username == cap.config.BASIC_AUTH_USERNAME \
            and password == cap.config.BASIC_AUTH_PASSWORD:
        return dict(username=username, password=password)


@jwtm.invalid_token_loader
def invalid_token_loader(mess):
    return dict(message=mess), httpcode.UNAUTHORIZED  # pragma: no cover


class RevokedTokenModel(db.Model):
    __tablename__ = 'revoked_tokens'

    id = db.Column(db.Integer, primary_key=True)
    jti = db.Column(db.String(120), nullable=False, unique=True)

    def __repr__(self):  # pragma: no cover
        return f"<RevokedToken: {self.id} - {self.jti}>"

    @classmethod
    def is_jti_blacklisted(cls, jti):  # pragma: no cover
        return bool(cls.query.filter_by(jti=jti).first())


@jwtm.token_in_blacklist_loader
def check_if_token_in_blacklist(decrypted_token):  # pragma: no cover
    return RevokedTokenModel.is_jti_blacklisted(decrypted_token['jti'])


class TokenHandler:
    revoked_token_model = RevokedTokenModel

    @classmethod
    def identity(cls):
        return jwt.get_jwt_identity()

    @classmethod
    def decode(cls, token):
        return ObjectDict(**jwt.decode_token(token))

    @classmethod
    def get_raw(cls):
        return ObjectDict(**jwt.get_raw_jwt())

    @classmethod
    def get_access(cls, identity=None, expires=None):
        return jwt.create_access_token(
            identity=identity or cls.identity(),
            expires_delta=timedelta(expires) if expires else None
        )

    @classmethod
    def get_refresh(cls, identity=None, expires=None):
        return jwt.create_refresh_token(
            identity=identity or cls.identity(),
            expires_delta=timedelta(expires) if expires else None
        )

    @classmethod
    def refresh(cls, expires=None):
        """

        :param expires: in seconds
        :return:
        """
        access_token = cls.get_access(expires=expires)
        decoded = cls.decode(access_token)
        return ObjectDict(
            access_token=access_token,
            expires_in=decoded.exp,
            issued_at=decoded.iat,
            token_type=cap.config.JWT_DEFAULT_TOKEN_TYPE,
            scope=cap.config.JWT_DEFAULT_SCOPE
        )

    @classmethod
    def create(cls, identity, refresh=True, expires_access=None, expires_refresh=None, scope=None):
        """

        :param identity: user identifier, generally the username
        :param refresh: enable refresh token
        :param expires_access: in seconds
        :param expires_refresh: in seconds
        :param scope:
        :return:
        """
        access_token = cls.get_access(identity=identity, expires=expires_access)
        decoded = cls.decode(access_token)
        resp = ObjectDict(
            access_token=access_token,
            expires_in=decoded.exp,
            issued_at=decoded.iat,
            token_type=cap.config.JWT_DEFAULT_TOKEN_TYPE,
            scope=scope or cap.config.JWT_DEFAULT_SCOPE
        )

        if refresh:
            resp.refresh_token = cls.get_refresh(identity=identity, expires=expires_refresh)

        return resp

    @classmethod
    def revoke(cls, token=None):
        token = cls.decode(token) if token else cls.get_raw()
        if token.jti:
            cls.revoked_token_model(jti=token.jti).add()

    @classmethod
    def dump(cls):
        return ObjectDict(
            token_type=cap.config.JWT_DEFAULT_TOKEN_TYPE,
            scope=cap.config.JWT_DEFAULT_SCOPE,
            **cls.get_raw()
        )



import functools

import flask
from webargs import fields, flaskparser

from flaskel import config, httpcode

parser = flaskparser.FlaskParser()
query = functools.partial(parser.use_args, location="query")
payload = functools.partial(parser.use_args, location="json")


class Fields:
    integer = fields.Integer
    string = fields.String
    decimal = fields.Decimal
    boolean = functools.partial(fields.Boolean, truthy={'true', 1}, falsy={'false', 0})
    positive = functools.partial(fields.Integer, validate=lambda x: x > 0)
    not_negative = functools.partial(fields.Integer, validate=lambda x: x >= 0)
    not_positive = functools.partial(fields.Integer, validate=lambda x: x <= 0)
    negative = functools.partial(fields.Integer, validate=lambda x: x < 0)
    isodate = functools.partial(fields.DateTime, format=config.DATE_ISO_FORMAT)
    list_of = functools.partial(fields.DelimitedList, cls_or_instance=string, delimiter='+')


# noinspection PyUnusedLocal
@parser.error_handler
def handle_error(error, *args, **kwargs):
    flask.abort(httpcode.BAD_REQUEST, response=error.messages)


def paginate(f=None):
    @query(dict(
        page=Fields.positive(missing=None),
        page_size=Fields.positive(missing=None),
        related=Fields.boolean(missing=False),
    ))
    def _route(*args, **kwargs):
        params = args[-1]
        if f is None:
            return params

        return f(*args[:-1], params=params, **kwargs)

    return _route if f is not None else _route()



import flask
from webargs import fields

from flaskel import httpcode
from flaskel.ext.auth import jwt, TokenHandler
from flaskel.utils.webargs import query
from . import auth


@auth.route('/token/access', methods=['POST'])
@query({
    'expire_access':  fields.Integer(missing=None, validate=lambda x: x > 0),
    'expire_refresh': fields.Integer(missing=None, validate=lambda x: x > 0),
})
def access_token(args):
    data = flask.request.json
    if data.email == 'email' and data.password == 'password':
        return TokenHandler.create(
            data.email,
            expires_access=args['expire_access'],
            expires_refresh=args['expire_refresh']
        )

    flask.abort(httpcode.UNAUTHORIZED)


@auth.route('/token/refresh', methods=['POST'])
@jwt.jwt_refresh_token_required
def refresh_token():
    return TokenHandler.refresh()


@auth.route('/token/check', methods=['GET'])
@jwt.jwt_required
def check_token():
    return TokenHandler.dump()



from .caching import Cache, caching
from .default import (
    builder,
    cfremote,
    cors,
    errors,
    logger,
    template
)
from .limit import FlaskIPBan, limiter, RateLimit

ip_ban = FlaskIPBan()

# { "name": (extension, parameters: dict), }
BASE_EXTENSIONS = {
    "cfremote": (cfremote,),  # MUST be the first
    "logger":   (logger,),  # MUST be the second
    "template": (template,),
    "builder":  (builder,),
    "errors":   (errors, {
        "dispatcher": 'subdomain',
        "response":   builder.on_accept(strict=False),
    }),
}



import psutil

from flaskel.http.batch import HTTPBatch
from flaskel.utils.datastruct import ObjectDict


def health_sqlalchemy(db, app, stm='SELECT 1'):
    """

    :param db:
    :param app:
    :param stm:
    :return:
    """
    try:
        with app.app_context():
            with db.engine.connect() as connection:
                connection.execute(stm)
    except Exception as exc:
        return False, str(exc)
    return True, None  # pragma: no cover


def health_mongo(db, app, cmd='ping'):
    """

    :param db:
    :param app:
    :param cmd:
    :return:
    """
    try:
        with app.app_context():
            db.db.command(cmd)
    except Exception as exc:
        return False, str(exc)
    return True, None  # pragma: no cover


def health_redis(db, app):
    """

    :param db:
    :param app:
    :return:
    """
    try:
        with app.app_context():
            db.ping()
    except Exception as exc:
        return False, str(exc)
    return True, None  # pragma: no cover


# noinspection PyUnusedLocal
def health_glances(conf=None, **kwargs):
    """

    :param conf:
    :return:
    """
    conf = conf() if callable(conf) else (conf or {})
    default_conf = ObjectDict(
        SYSTEM_ENDPOINT='http://localhost:4000/api/2',
        SYSTEM_CPU_THRESHOLD=90,
        SYSTEM_MEM_THRESHOLD=90,
        SYSTEM_FS_THRESHOLD=85,
        SYSTEM_MEM_INCLUDE_SWAP=True,
        SYSTEM_FS_MOUNT_POINTS=('/',),
        SYSTEM_DUMP_ALL=False
    )
    conf = ObjectDict(**{**default_conf, **conf})
    resp = ObjectDict(errors=[])
    units = ['cpu', 'mem', 'fs']
    if conf.SYSTEM_MEM_INCLUDE_SWAP:
        units.append('memswap')

    th_mem = conf.SYSTEM_MEM_THRESHOLD
    th_cpu = conf.SYSTEM_CPU_THRESHOLD
    th_fs = conf.SYSTEM_FS_THRESHOLD

    responses = HTTPBatch().request([dict(url=f"{conf.SYSTEM_ENDPOINT}/{u}") for u in units])
    for i, r in enumerate(responses):
        if r.exception:
            return False, str(r.exception)
        resp[units[i]] = r.body

    if resp.cpu.total > th_cpu:
        resp.errors.append(f"high CPU usage: {resp.cpu.total}, threshold: {th_cpu}")

    if resp.mem.percent > th_mem:
        resp.errors.append(f"high RAM usage: {resp.mem.percent}, threshold: {th_mem}")
    if conf.SYSTEM_MEM_INCLUDE_SWAP:
        if resp.memswap.percent > th_mem:
            resp.errors.append(f"high SWAP usage: {resp.memswap.percent}, threshold: {th_mem}")

    for f in resp.fs:
        if f.mnt_point in conf.SYSTEM_FS_MOUNT_POINTS and f.percent > th_fs:
            resp.errors.append(f"high DISK usage on {f.mnt_point}: {f.percent}, threshold: {th_fs}")

    output = resp if conf.SYSTEM_DUMP_ALL else (resp.errors or None)
    return bool(not resp.errors), dict(messages=output)


# noinspection PyProtectedMember,PyUnusedLocal
def health_system(conf=None, **kwargs):
    """

    :param conf:
    :return:
    """
    conf = conf() if callable(conf) else (conf or {})
    default_conf = dict(
        SYSTEM_FS_MOUNT_POINTS=('/',),
        SYSTEM_CPU_THRESHOLD=90,
        SYSTEM_MEM_THRESHOLD=90,
        SYSTEM_FS_THRESHOLD=85,
        SYSTEM_MEM_INCLUDE_SWAP=True,
        SYSTEM_DUMP_ALL=False
    )
    conf = ObjectDict(**{**default_conf, **conf})
    resp = ObjectDict(errors=[])
    th_mem = conf.SYSTEM_MEM_THRESHOLD
    th_cpu = conf.SYSTEM_CPU_THRESHOLD
    th_fs = conf.SYSTEM_FS_THRESHOLD

    resp.mem = ObjectDict(**psutil.virtual_memory()._asdict())
    if resp.mem.percent > th_mem:
        resp.errors.append(f"high RAM usage: {resp.mem.percent}, threshold: {th_mem}")
    if conf.SYSTEM_MEM_INCLUDE_SWAP:
        resp.swap = ObjectDict(**psutil.swap_memory()._asdict())
        if resp.swap.percent > th_mem:
            resp.errors.append(f"high SWAP usage: {resp.swap.percent}, threshold: {th_mem}")

    resp.cpu = psutil.cpu_percent()
    if resp.cpu > th_cpu:
        resp.errors.append(f"high CPU usage: {resp.cpu}, threshold: {th_cpu}")

    resp.disk = ObjectDict()
    for f in conf.SYSTEM_FS_MOUNT_POINTS:
        resp.disk[f] = ObjectDict(**psutil.disk_usage(f)._asdict())
        percent = resp.disk[f].percent
        if percent > th_mem:
            resp.errors.append(f"high DISK usage on '{f}': {percent}, threshold: {th_fs}")

    output = resp if conf.SYSTEM_DUMP_ALL else (resp.errors or None)
    return bool(not resp.errors), dict(messages=output)



import logging

from flask_apscheduler import APScheduler
from flaskel.utils.uuid import get_uuid


class APJobs(APScheduler):
    def init_app(self, app):
        super().init_app(app)

        if app.debug:
            logger = logging.getLogger('apscheduler')
            logger.setLevel(logging.DEBUG)
            logger = logging.getLogger('flask_apscheduler')
            logger.setLevel(logging.DEBUG)

        app.config.setdefault('SCHEDULER_AUTO_START', False)

        if not hasattr(app, 'extensions'):
            app.extensions = dict()  # pragma: no cover
        app.extensions['scheduler'] = self

        if app.config.SCHEDULER_AUTO_START is True:
            self.start()

    def add_job(self, func, kwargs=None, **kw):
        """

        :param func:
        :param kwargs:
        :return:
        """
        kwargs = kwargs or {}
        kwargs['app'] = self.app
        kw['kwargs'] = kwargs
        return super().add_job(id=get_uuid(), func=func, **kw)


scheduler = APJobs()



# example api blueprint
#
from flask import Blueprint

from flaskel.ext import cors, errors

api = Blueprint(
    'api', __name__,
    subdomain='api',
    static_folder="static",
    static_url_path="/cdn/"
)

cors.init_app(api)
errors.api_register(api)



from flask import Blueprint

from flaskel.ext import cors, errors
from flaskel.views import JSONRPCView, TransparentProxyView, ConfProxyView
from tests.blueprints.api.resource import APIResource
from tests.blueprints.api.rpc import MyJsonRPC

api = Blueprint('api', __name__, subdomain='api')

cors.init_app(api)
errors.api_register(api)

APIResource.register(api, 'resource_api', '/resources')

jsonRPCView = JSONRPCView()
jsonRPCView.load_from_object(MyJsonRPC())
jsonRPCView.register(api, 'myJsonRPC', "/rpc")


class CustomProxy(TransparentProxyView):
    methods = ['POST']


CustomProxy.register(
    api, 'proxyview', '/proxy',
    host='https://httpbin.org', method='POST', url='/anything'
)

ConfProxyView.register(api, 'confproxy', '/confproxy', config_key='PROXIES.CONF')



import flask

from flaskel import cap, httpcode
from .base import BaseView


class ApiDocTemplate(BaseView):
    apispec_view = 'apispec'

    def dispatch_request(self):
        if not cap.config.APIDOCS_ENABLED:
            flask.abort(httpcode.NOT_FOUND)

        rapidoc_version = cap.config.RAPIDOC_VERSION or '8.4.3'
        rapidoc_theme = cap.config.RAPIDOC_THEME or 'dark'
        page_title = f"{cap.config.APP_NAME} - API DOCS"

        spec_url = flask.url_for(
            self.apispec_view,
            _external=True,
            _scheme=cap.config.PREFERRED_URL_SCHEME
        )

        return f"""
            <!doctype html>
            <html>
            <head>
            <meta charset="utf-8"/>
            <meta http-equiv="X-UA-Compatible" content="IE=edge,chrome=1"/>
            <meta name="viewport" content="width=device-width,initial-scale=1,shrink-to-fit=no"/>
            <title>{page_title}</title>
            <script type="module"
                src="https://unpkg.com/rapidoc@{rapidoc_version}/dist/rapidoc-min.js">
            </script>
            </head>
            <body>
                <rapi-doc theme="{rapidoc_theme}" spec-url="{spec_url}"></rapi-doc>
            </body>
            </html>
        """


class ApiSpecTemplate(BaseView):
    api_version = '1.0.0'
    url_version = '/v1'

    def dispatch_request(self):
        if not cap.config.APIDOCS_ENABLED:
            flask.abort(httpcode.NOT_FOUND)

        data = cap.config.APISPEC
        scheme = cap.config.PREFERRED_URL_SCHEME or 'http'
        current_server = f"{scheme}://{cap.config.SERVER_NAME}/{self.url_version}"

        try:
            data['info']['version'] = self.api_version
            variables = data['servers'][0]['variables']
            variables['version']['default'] = self.url_version
            variables['host']['default'] = current_server
        except (AttributeError, IndexError, KeyError, TypeError) as exc:
            cap.logger.debug(str(exc))

        return data
